class Buffer{
  private char[] buf;               // storage
  private int last = 0;             // last occupied position
  private int writers_waiting = 0;  // # of threads waiting in put()
  private int readers_waiting = 0;  // # of threads waiting in get()
  public Buffer (int sz) {
        buf = new char[sz];
}
  public boolean isFull()  { return (last == buf.length); }
  public boolean isEmpty() { return (last == 0);          }

  public synchronized void put(char c) {
        Thread t = Thread.currentThread();

        while(isFull()) {
          try     { writers_waiting++;  wait(); }
          catch   (InterruptedException e) { }
          finally { writers_waiting--; }
        }
        System.out.println("Writer: " + t + " put at " + last);
        buf[last++] = c;
        if (readers_waiting > 0) {
          System.out.println("Writer: " + t + " notifies all the readers");
          notifyAll();
        }
  }
  public synchronized char get() {
  		Thread t = Thread.currentThread();
        while(isEmpty()) {
          try     { readers_waiting++;  wait(); }
          catch   (InterruptedException e) { }
          finally { readers_waiting--; }
        }
        char c =  buf[0];
        System.out.println("Reader : " + t + " got number (last="+(last-1)+")");
        System.arraycopy(buf, 1, buf, 0, --last);
        if (writers_waiting > 0) {
          System.out.println("Reader: " + t + " notifies all the writers");
          notifyAll();
        }
        return c;
  }
}
class Writer implements Runnable {
  Buffer buf;
  Writer(Buffer buf,String name) {
    this.buf = buf;
    new Thread(this, name).start();
  }
  public void run() {
    char c = 0;

    while(true) {
      try{
	    Thread.sleep(1000);
	    }catch(InterruptedException e){}

      buf.put(c++);
      // System.out.println("Writer put : " + (int)c);
      }
  }
}
class Reader implements Runnable {
  Buffer buf;
  Reader(Buffer buf,String name) {
    this.buf = buf;
    new Thread(this, name).start();
  }
  public void run() {
    char c;

    while(true) {
      try{
	    Thread.sleep(2000);
	    }catch(InterruptedException e){}

      c=buf.get();
      // System.out.println("Reader got : " + (int) c);
    }
  }
}
class ReadersWriters {
  public static void main(String args[]) {
    Buffer buf = new Buffer(10);
    new Writer(buf,"writer1");
    new Reader(buf,"reader1");
    new Writer(buf,"writer2");
  }
}